# Copyright 2014-2015 Nathan West
#
# This file is part of autocommand.
#
# autocommand is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# autocommand is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with autocommand.  If not, see <http://www.gnu.org/licenses/>.

import sys
from inspect import signature, Parameter, getdoc
from argparse import ArgumentParser, _StoreConstAction
from contextlib import contextmanager
from io import IOBase


__version__ = '2.0.1'


_empty = Parameter.empty


class AutocommandError(TypeError):
    '''Base class for autocommand exceptions'''


class AnnotationError(AutocommandError):
    '''Annotation error: annotation must be a string, type, or tuple of both'''


class PositionalArgError(AutocommandError):
    '''
    Postional Arg Error: autocommand can't handle postional-only parameters
    '''


class KWArgError(AutocommandError):
    '''kwarg Error: autocommand can't handle a **kwargs parameter'''


def _get_type_description(annotation):
    '''
    Given an annotation, return the (type, description) for the parameter
    '''
    if annotation is _empty:
        return None, None
    elif isinstance(annotation, type):
        return annotation, None
    elif isinstance(annotation, str):
        return None, annotation
    elif isinstance(annotation, tuple):
        arg1, arg2 = annotation
        if isinstance(arg1, type) and isinstance(arg2, str):
            return arg1, arg2
        elif isinstance(arg1, str) and isinstance(arg2, type):
            return arg2, arg1

    raise AnnotationError(annotation)


def _make_argument(param, used_char_args):
    '''
    Get the *args and **kwargs to use for parser.add_argument for a given
    parameter. used_char_args is the set of -short options currently already in
    use.
    '''
    if param.kind is param.POSITIONAL_ONLY:
        raise PositionalArgError(param)
    elif param.kind is param.VAR_KEYWORD:
        raise KWArgError(param)

    arg_spec = {}
    is_option = False

    # Get the type and default from the annotation.
    arg_type, description = _get_type_description(param.annotation)

    # Get the default value
    default = param.default

    # If there is no explicit type, and the default is present and not None,
    # infer the type from the default.
    if arg_type is None and default not in (_empty, None):
        arg_type = type(default)

    # Add default. The presence of a default means this is an option, not an
    # argument.
    if default is not _empty:
        arg_spec['default'] = default
        is_option = True

    # Add the type
    if arg_type is not None:
        # Special case for bool: make it just a --switch
        if arg_type is bool:
            if not default or default is _empty:
                arg_spec['action'] = 'store_true'
            else:
                arg_spec['action'] = 'store_false'

            # Switches are always options
            is_option = True

        # Special case for file types: make it a string type, for filename
        elif isinstance(default, IOBase):
            arg_spec['type'] = str

        # TODO: special case for list type.
        #   - How to specificy type of list members?
        #       - param: [int]
        #       - param: int =[]
        #   - action='append' vs nargs='*'

        else:
            arg_spec['type'] = arg_type

    # nargs: if the signature includes *args, collect them as trailing CLI
    # arguments in a list. *args can't have a default value, so it can never be
    # an option.
    if param.kind is param.VAR_POSITIONAL:
        # TODO: consider depluralizing metavar/name here.
        arg_spec['nargs'] = '*'

    # Add description.
    if description is not None:
        arg_spec['help'] = description

    # Get the --flags
    flags = []
    name = param.name

    if is_option:
        # Add the first letter as a -short option.
        for letter in name[0], name[0].swapcase():
            if letter not in used_char_args:
                used_char_args.add(letter)
                flags.append('-{}'.format(letter))
                break

        # If the parameter is a --long option, or is a -short option that
        # somehow failed to get a flag, add it.
        if len(name) > 1 or not flags:
            flags.append('--{}'.format(name))

        arg_spec['dest'] = name
    else:
        flags.append(name)

    return flags, arg_spec


def _make_parser(main_sig, description, epilog, add_nos):
    '''
    Given the signature of a function, create an ArgumentParser
    '''
    parser = ArgumentParser(description=description, epilog=epilog)

    used_char_args = {'h'}
    # Add each argument. Do single-character arguments first, if
    # present, so that they get priority, and don't have to get --long
    # versions. sorted is stable, so the parameters will still be in
    # relative order
    for param in sorted(
            main_sig.parameters.values(),
            key=lambda param: len(param.name) > 1):
        flags, spec = _make_argument(param, used_char_args)
        action = parser.add_argument(*flags, **spec)

        # If requested, add --no- option counterparts. Because the
        # option/argument names can't have a hyphen character, these
        # shouldn't conflict with any existing options.
        # TODO: decide if it's better, stylistically, to do these at
        # the end, AFTER all of the parameters.
        if add_nos and isinstance(action, _StoreConstAction):
            parser.add_argument(
                '--no-{}'.format(action.dest),
                action='store_const',
                dest=action.dest,
                const=action.default)
            # No need for a default=, as the first action takes
            # precedence.
    return parser


def autocommand(
        module=None, *,
        description=None,
        epilog=None,
        add_nos=False,
        parser=None):
    '''
    Decorator to create an autocommand function. The function's signature is
    analyzed, and an ArgumentParser is created, using the `description` and
    `epilog` parameters, to parse command line arguments corrosponding to the
    function's parameters. The function's signature is changed to accept an
    argv parameter, as from `sys.argv[1:]`, though you can supply your own.
    When called, the function parses the arguments provided, then supplies them
    to the decorated function. Keep in mind that this happens with plain
    argparse, so supplying invalid arguments or '-h' will cause a usage
    statement to be printed and a `SystemExit` to be raised.

    Optionally, pass a module name (typically `__name__`) as the first argument
    to `autocommand`. If you do, and it is "__main__" or True, the decorated
    function is called immediately with `sys.argv[1:]`, and the progam is
    exited with the return value; this is so that you can call
    `@autocommand(__name__)` and still be able to import the module for
    testing.

    The function can also be called with no arguments; in this case,
    `sys.argv[1:]` is used by default. This is so that an autocommand function
    can be used as a setuptools entry point, as well as a normal main function.

    The `desctiption` and `epilog` parameters corrospond to the same respective
    argparse parameters. If no description is given, it defaults to the
    decorated functions's docstring, if present.

    If add_nos is True, every boolean option will have a --no- version created
    as well, which inverts the option. For instance, the --verbose option will
    have a --no-verbose counterpart. These are not mutually exclusive-
    whichever one appears last on the command line will have precedence.

    If a parser is given, it is used instead of one generated from the function
    signature.

    The decorated function is attached to the result as the `main` attribute,
    and the parser is attached as the `parser` attribute.
    '''

    # If @autocommand is used instead of @autocommand(__name__)
    if callable(module):
        return autocommand(
            description=description,
            epilog=epilog,
            add_nos=add_nos,
            parser=parser)(module)

    def decorator(main):
        main_sig = signature(main)
        local_parser = parser or _make_parser(
            main_sig, description or getdoc(main), epilog, add_nos)

        def main_wrapper(argv=None):
            # Import here, rather than setting the parameter to a default, so
            # that changes to argv (however unlikely) are reflected in future
            # calls.
            if argv is None:
                argv = sys.argv[1:]

            # Get empty argument binding, to fill with parsed arguments. This
            # object does all the heavy lifting of turning named arguments into
            # into correctly bound *args and **kwargs.
            func_args = main_sig.bind_partial()
            func_args.arguments.update(vars(local_parser.parse_args(argv)))

            return main(*func_args.args, **func_args.kwargs)

        # If we are running as a script/program, call main right away and exit.
        if module == '__main__' or module is True:
            sys.exit(main_wrapper())

        # Otherwise, attach the wrapped main function and parser, and return
        # the wrapper.
        main_wrapper.main = main
        main_wrapper.parser = local_parser
        return main_wrapper

    return decorator


@contextmanager
def smart_open(filename_or_file, *args, **kwargs):
    '''
    This context manager allows you to open a filename, if you want to default
    some already-existing file object, like sys.stdout, which shouldn't be
    closed at the end of the context. If the filename argument is a str, bytes,
    or int, the file object is created via a call to open with the given *args
    and **kwargs, sent to the context, and closed at the end of the context,
    just like "with open(filename) as f:". If it isn't one of the openable
    types, the object simply sent to the context unchanged. Example:

        def work_with_file(name=sys.stdout):
            with smart_open(name) as f:
                # Works correctly if name is a str filename or sys.stdout
                print("Some stuff", file=f)
                # If it was a filename, f is closed at the end here.
    '''
    try:
        file = open(filename_or_file, *args, **kwargs)
    except TypeError:
        yield filename_or_file
    else:
        with file:
            yield file



from unittest.mock import patch, sentinel
import pytest
from argparse import ArgumentParser
from autocommand import autocommand


@pytest.mark.parametrize('module_name', ['__main__', True])
def test_name_main(module_name):
    '''
    Test that applying the autocommand decorator to a function with '__name__'
    or True as the argument causes it to be called with sys.argv, and then
    exited with the return value of the function.
    '''

    with patch('sys.argv', ['prog', 'foo', '0']):
        try:
            @autocommand(module_name)
            def prog(arg1, arg2: int):
                assert arg1 == 'foo'
                assert arg2 == 0
                return sentinel.exit_code

        except SystemExit as e:
            assert e.code is sentinel.exit_code
        else:
            pytest.fail("autocommand function didn't exit")


def test_no_args():
    with patch('sys.argv', ['prog', 'foo', '0']):
        @autocommand
        def prog(arg1, arg2: int):
            assert arg1 == 'foo'
            assert arg2 == 0
            return sentinel.exit_code

        assert prog() == sentinel.exit_code


def test_main_attr():
    @autocommand
    def prog(arg1=False):
        return arg1

    func = prog.main

    assert func() == False
    assert func(True) == True
    assert func(arg1=10) == 10
    assert prog([]) == False
    assert prog(['--arg1']) == True


def test_docstring(capsys):
    @autocommand
    def prog():
        '''
        This is the
        help text
        '''

    with pytest.raises(SystemExit):
        prog(['-h'])

    out, err = capsys.readouterr()
    assert "This is the help text" in out
    assert prog.parser.description == "This is the\nhelp text"


def test_desc_epilog(capsys):
    @autocommand(
        description="This is the description",
        epilog="This is the epilog")
    def prog():
        pass

    with pytest.raises(SystemExit):
        prog(['-h'])

    out, err = capsys.readouterr()
    assert "This is the description" in out
    assert "This is the epilog" in out
    assert prog.parser.description == "This is the description"
    assert prog.parser.epilog == "This is the epilog"


def test_custom_parser():
    '''
    Test a custom parser. The parser includes features that can't be created by
    autocommand, but should autocommand should still handle being able to parse
    and pass to the function.
    '''
    parser = ArgumentParser()
    parser.add_argument('arg', nargs='?')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-v', '--verbose', action='store_true')
    group.add_argument('-q', '--quiet', action='store_true')

    @autocommand(parser=parser)
    def prog(arg, verbose, quiet):
        return arg, verbose, quiet

    assert prog([]) == (None, False, False)
    assert prog(['foo']) == ('foo', False, False)
    assert prog(['-v']) == (None, True, False)
    assert prog(['-q']) == (None, False, True)

    with pytest.raises(SystemExit):
        prog(['-v', '-q'])


def test_nos():
    @autocommand(add_nos=True)
    def prog(verbose=False):
        return verbose

    assert prog([]) == False
    assert prog(['--verbose']) == True
    assert prog(['--no-verbose']) == False
    assert prog(['--verbose', '--no-verbose']) == False
    assert prog(['--no-verbose', '--verbose']) == True



from argparse import ArgumentParser
from inspect import Parameter
from pytest import mark
from unittest.mock import Mock
from io import IOBase

from autocommand import _make_argument


def param(**kwargs):
    '''
    Create a inspect.Parameter with the (default) kind POSITIONAL_OR_KEYWORD
    '''
    kwargs.setdefault('kind', Parameter.POSITIONAL_OR_KEYWORD)
    return Parameter(**kwargs)

fakefile = Mock(IOBase)

arg_cases = [
    # Basic argument
    (param(name='arg'),
        {}),
    # Typed argument
    (param(name='arg', annotation=int),
        {'type': int}),

    # Basic option
    (param(name='arg', default=None),
        {'default': None, 'dest': 'arg'}),

    # Bool argument
    (param(name='arg', annotation=bool),
        {'action': 'store_true', 'dest': 'arg'}),
    # True option
    (param(name='arg', default=True),
        {'action': 'store_false', 'dest': 'arg', 'default': True}),
    # False option
    (param(name='arg', default=False),
        {'action': 'store_true', 'dest': 'arg', 'default': False}),

    # Implicitly typed default option
    (param(name='arg', default=0),
        {'default': 0, 'dest': 'arg', 'type': int}),

    # File type
    (param(name='arg', default=fakefile),
        {'default': fakefile, 'dest': 'arg', 'type': str}),

    # Description
    (param(name='arg', annotation="description text"),
        {'help': "description text"}),

    # Description + type
    (param(name='arg', annotation=("description text", int)),
        {'help': "description text", "type": int}),

    # Positional
    (param(name='args', kind=Parameter.VAR_POSITIONAL),
        {'nargs': '*'}),

    # Typed positional
    (param(name='args', kind=Parameter.VAR_POSITIONAL, annotation=int),
        {'nargs': '*', 'type': int}),
]


@mark.parametrize('param, kwargs', arg_cases)
def test_make_arguments(param, kwargs):
    made_args, made_kwargs = _make_argument(param, set())

    assert made_kwargs == kwargs

flag_cases = [
    # Basic option
    (param(name='arg', default=None), [],
        ['-a', '--arg'], {'a'}),

    # Cased option
    (param(name='arg', default=None), ['a'],
        ['-A', '--arg'], {'a', 'A'}),

    # Downcased option
    (param(name='ARG', default=None), ['A'],
        ['-a', '--ARG'], {'a', 'A'}),

    # Unavailable option
    (param(name='arg', default=None), ['a', 'A'],
        ['--arg'], {'a', 'A'}),

    # Single char
    (param(name='a', default=None), [],
        ['-a'], {'a'}),

    # Cased char
    (param(name='a', default=None), ['a'],
        ['-A'], {'a', 'A'}),

    # Unavailable char
    (param(name='a', default=None), ['a', 'A'],
        ['--a'], {'a', 'A'})
]


@mark.parametrize('param, used, flags, new_used', flag_cases)
def test_flag_creation(param, used, flags, new_used):
    # Ensure transitivity: copy the set
    used = set(used)
    made_args, made_kwargs = _make_argument(param, used)

    assert made_args == flags
    assert used == new_used



from contextlib import contextmanager
from autocommand import autocommand
import pytest
from pytest import raises
from functools import wraps


def run_autocmd_tests(*cases):
    '''
    This decorator wraps up a bunch of test cases for an autocommand function.
    It basically works like this:

    @run_autcommand_tests(
        ('--arg 1',   1),
        ('-a 1',      1)
        ('--arg foo', SystemExit))
    def test_int_option(arg=0):
        return arg

    The test_int_option is internally converted into an autocommand function,
    then the function is turned into a test case. The test case runs each set
    of options (which have .split() called on them) on the autocommand and
    checks that the return value matches. If `SystemExit` is used, the program
    is checked instead that it raises a SystemExit.
    '''
    def decorator(func):
        def test_wrapper():
            prog = autocommand(func)
            for args, result in cases:
                args = args.split()
                if result is SystemExit:
                    with raises(SystemExit):
                        prog(args)
                else:
                    assert prog(args) == result
        return test_wrapper
    return decorator


@run_autocmd_tests(
    ('', 0),
    ('arg', SystemExit))
def test_no_args():
    return 0


@run_autocmd_tests(
    ('', SystemExit),
    ('arg', 'arg'))
def test_single_arg(arg1):
    return arg1


@run_autocmd_tests(
    ('arg', SystemExit),
    ('1', 1))
def test_int_arg(arg1: int):
    return arg1


@run_autocmd_tests(
    ('',        []),
    ('foo',     ['foo']),
    ('foo bar', ['foo', 'bar']))
def test_var_args(*args):
    return list(args)


@run_autocmd_tests(
    ('',    []),
    ('foo', SystemExit),
    ('1',   [1]),
    ('1 2', [1, 2]))
def test_typed_var_args(*args: int):
    return list(args)


@run_autocmd_tests(
    ('',          None),
    ('--arg foo', 'foo'),
    ('-a foo',    'foo'),
    ('--arg=foo', 'foo'),
    ('-afoo',     'foo'))
def test_simple_option(arg=None):
    return arg


@run_autocmd_tests(
    ('',          None),
    ('--arg foo', SystemExit),
    ('--arg 1',   1))
def test_typed_int_option(arg: int =None):
    return arg



import autocommand
from inspect import Parameter, Signature
from unittest.mock import patch
from pytest import raises, mark


# Note: The (int, 1) case is to provide 100% branch coverage
@mark.parametrize('annotation', [1, (int, 1)])
def test_annotation_error(annotation):
    with raises(autocommand.AnnotationError):
        @autocommand.autocommand
        def annotate(arg1: annotation):
            pass


def test_kwarg_error():
    with raises(autocommand.KWArgError):
        @autocommand.autocommand
        def kwarg(**kwargs):
            pass


def test_positional_error():
    fake_parameter = Parameter(name='arg', kind=Parameter.POSITIONAL_ONLY)
    fake_sig = Signature([fake_parameter])

    # autocommand uses from ... import ...
    with patch('autocommand.signature', return_value=fake_sig):
        with raises(autocommand.PositionalArgError):
            @autocommand.autocommand
            def positional(arg):
                pass



import autocommand
from inspect import Parameter
import pytest


@pytest.mark.parametrize('annotation, type, description', [
    (Parameter.empty,      None, None),
    (int,                  int, None),
    ('description',        None, 'description'),
    ((int, 'description'), int, 'description'),
    (('description', int), int, 'description')])
def test_annotation(annotation, type, description):
    assert autocommand._get_type_description(annotation) == (type, description)


def test_invalid_annotation():
    with pytest.raises(autocommand.AnnotationError):
        autocommand._get_type_description(None)



from autocommand import smart_open
from pytest import raises, fixture, yield_fixture
from unittest.mock import patch, MagicMock
from io import IOBase


def mockfile():
    return MagicMock(IOBase)


def patch_open(*args, **kwargs):
    return patch('builtins.open', *args, return_value=mockfile(), **kwargs)


def test_path():
    '''
    Test smart_open with a file path. It the file should be opened, then closed
    at the end of the context.
    '''
    with patch_open() as mock_open:
        with smart_open('/path/to/file') as file:
            file.readline()

    mock_open.assert_called_once_with('/path/to/file')
    file.__enter__.assert_called_once_with()
    file.readline.assert_called_once_with()
    file.__exit__.assert_called_once_with(None, None, None)


def test_file():
    '''
    Test smart_open with NOT a file path. The input object should be returned,
    and no context methods should be called on it
    '''
    fakefile = mockfile()
    # For some reason open(MagicMock()) doesn't throw, so we have to mock open

    with patch_open(side_effect=TypeError):
        with smart_open(fakefile) as file:
            file.readline()

    assert file is fakefile
    assert not file.__enter__.called
    file.readline.assert_called_once_with()
    assert not file.__enter__.called


def test_open_args(tmpdir):
    '''
    Test that smart_open can forward kwargs correctly to open.
    '''
    with patch_open() as mock_open:
        with smart_open('/path/to/file', 'w', encoding='utf8') as file:
            pass

    mock_open.assert_called_once_with('/path/to/file', 'w', encoding='utf8')
